#ifndef DE_H
#define DE_H

#include <iostream>
#include <fstream>
#include "chaos.h"

#define stBest1			0
#define stRand1			1
#define stRandToBest1	2
#define stBest2			3
#define stRand2			4

#define stExp           0
#define stBin           1

using namespace std;

class DE
{
    public:
        /*konstruktor, destruktor*/
        DE(double CR, int NP, double F, int generations, int D);
        virtual ~DE();

        /*gettery*/
        int getPopSize() {return popSize;}
        int getDim() {return dimension;}
        void getPopulation(int generation); //vypise populaci dane generace - fce pouze pro ucely testovani
        void getWholePop(int generation);   //vypise celou generaci - fce pro vypracovani grafu vyvoje jedincu

        /*settery*/
        void init(chaos generator, double mini, double maxi, int testFnctn);  //inicializacni funkce.
        void setPopulation();           //zalozi populaci.

        /*zakladni funkce diferencialniho evolucniho algoritmu*/
        void evolve(int mStrategy, int cStrategy);  //evolucni cyklus
        void binCross();            //binarni krizeni
        void expCross(int i);       //exponencialni krizeni

        /*mutacni algoritmy*/
        void best1(int i);
        void rand1(int i);
        void randToBest1(int i);
        void best2(int i);
        void rand2(int i);

        int selectNoiseIndex();     //fce, ktera urci u exp. krizeni, kolik parametru vezmeme ze sumoveho vektoru

        /*cost function*/
        double costFunction(double * testPopulation, int index, int populationSize);

        double chaosRand(double minimum, double maximum);

    protected:
        double * population;//dynamicke pole populace (matice NP x D, kde jedinec = sloupec)
        double * newPop;    //nove vznikajici populace
        double * noise;     //sumovy vektor vznikly mutaci populace
        double * trial;     //zkusebni vektor vznikly krizenim sumoveho vektoru a 4. rodice
        double * best;      //nejlepsi dosavadni reseni problematiky

        void selectParents(int *r1, int *r2=0, int *r3=0, int *r4=0, int *r5=0);//pomocna funkce pro nahodny vyber rodicu

    private:
        /* ZAKLADNI PARAMETRY ALGORITMU DIFERENCIALNI EVOLUCE */
        double crossover;   //CR - prah krizeni z intervalu <0;1>. Pri nabyti krajnich hodnot: 0 - kopie ctvrteho rodice; 1 - krizeni 3 rodicu, 4. vynechan
        int popSize;        //NP - velikost populace. <2D, 100D>. Nemelo by byt mensi nez 4.
        double factor;      //F - mutacni konstanta. <0, 2>
        int generations;    //>0 pocet evolucnich cyklu, behem nichz se populace vyviji
        int dimension;      //D - dimenze problemu - pocet argumentu ucelove funkce.
        /* KONEC ZAKLADNICH PARAMETRU DIFERENCIALNI EVOLUCE */

        int testFunction;   //vybrana testovaci funkce
        double minimum, maximum;
        int activeParent;   //index aktivniho rodice
        double bestCV;      //nejlepsi hodnota Cost Value



        int stagnation;     //priznak stagnace
        int maxStagnation;  //max priznak stagnace

        ofstream myfile;   //zde zapisujeme vysledky
        chaos generator;    //generator chaosu
};

#endif // DE_H
