#include <iostream>
#include "include/DE.h"
#include "include/chaos.h"
#include <fstream>
#include <time.h>

#define PI  3.14159265

#define stDeJong	0
#define ndDeJong	1
#define rdDeJong	2
#define thDeJong 	3
#define Rastrigin	4
#define Schwefel	5
#define Griewangk	6
#define SinEnvelope	7
#define StretchVSin 8
#define Ackley 		9
#define ndAckley	10
#define Eggplate 	11
#define Rana		12
#define Patological	13
#define Michalewicz	14
#define Masters		15

#define stBest1			0
#define stRand1			1
#define stRandToBest1	2
#define stBest2			3
#define stRand2			4

#define stExp           0
#define stBin           1

using namespace std;

double minimum, maximum;

/* vymezeni intervalu hodnot jednotlivych testovacich funkci*/
void setBoundaries(int testFunction, bool mini, bool maxi)
{
    double mi, ma;

    if (mini){  // mini==true - minimum je zadano v konfigu
        mi=minimum;
    }
    if (maxi){
        ma=maximum;
    }


    switch(testFunction)
    {
    case stDeJong:
    case Masters:
        minimum=-5;
        maximum=5;
        break;
    case ndDeJong:
    case rdDeJong:
        minimum=-2;
        maximum=2;
        break;
    case thDeJong:
        minimum=-1.28;
        maximum=1.28;
        break;
    case Rastrigin:
        minimum=-5.12;
        maximum=5.12;
        break;
    case Schwefel:
    case Eggplate:
    case Rana:
        minimum=-512;
        maximum=512;
        break;
    case Griewangk:
        minimum=-600;
        maximum=600;
        break;
    case SinEnvelope:
    case Patological:
        minimum=-100;
        maximum=100;
        break;
    case StretchVSin:
        minimum=-10;
        maximum=10;
        break;
    case Ackley:
    case ndAckley:
        minimum=-30;
        maximum=30;
        break;
    case Michalewicz:
        minimum=0;
        maximum=PI;
        break;
    }

    if (mini){
        minimum=mi;
    }
    if (maxi){
        maximum=ma;
    }

}

int main()
{
    cout << "Algoritmus diferencialni evoluce s prvky deterministickeho chaosu" << endl << endl;

    string lineBuf;
    string optionBuf;
    ifstream confFile( "config.txt" );

    int testFunction, mStrategy, cStrategy, NP, generations, D, ch_system;
    double CR, F;

    /*Nacteni konfiguracniho souboru*/
    if ( confFile.is_open() )
    {
        while ( getline( confFile, lineBuf ) )
        {
            optionBuf = "Testovaci funkce = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                testFunction = atoi( lineBuf.c_str() );
                cout << optionBuf << testFunction << "\t";
            }

            optionBuf = "Mutacni strategie = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                mStrategy = atoi( lineBuf.c_str() );
                cout << optionBuf << mStrategy << "\t";
            }

            optionBuf = "Strategie krizeni = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                cStrategy = atoi( lineBuf.c_str() );
                cout << optionBuf << cStrategy << endl;
            }

            optionBuf = "Minimum = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                if (lineBuf == "x")
                    setBoundaries(testFunction, false, false);    //vymezeni defaultniho intervalu setreni dane testovaci funkce
                else
                minimum = atof( lineBuf.c_str() );
                cout << optionBuf << minimum << "\t\t";
            }

            optionBuf = "Maximum = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                if (lineBuf == "x")
                    setBoundaries(testFunction, true, false);    //vymezeni defaultniho intervalu setreni dane testovaci funkce
                else
                    maximum = atof( lineBuf.c_str() );
                cout << optionBuf << maximum << "\t\t";
            }

            optionBuf = "CR = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                CR = atof( lineBuf.c_str() );
                cout << optionBuf << CR << endl;
            }

            optionBuf = "NP = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                NP = atoi( lineBuf.c_str() );
                cout << optionBuf << NP << "\t\t\t";
            }

            optionBuf = "F = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                F = atof( lineBuf.c_str() );
                cout << optionBuf << F << endl;
            }

            optionBuf = "D = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                D = atoi( lineBuf.c_str() );
                cout << optionBuf << D << "\t\t\t";
            }

            optionBuf = "generace = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                generations = atoi( lineBuf.c_str() );
                cout << optionBuf << generations << "\t\t";
            }

            optionBuf = "Chaoticky system = ";
            if ( ( ( int )lineBuf.find( optionBuf ) ) != -1 )
            {
                lineBuf.erase( 0, optionBuf.length() );
                ch_system = atoi( lineBuf.c_str() );
                cout << optionBuf << ch_system << endl << endl;
            }
        }
        cout << "Inicializace z konfiguracniho souboru probehla uspesne." << endl;
        confFile.close();
    }
    else //pokud nemuzeme nacist konfiguracni soubor, zadame defaultni parametry
    {
        testFunction=stDeJong;
        mStrategy=stRand1;
        cStrategy=stBin;
        NP=20;
        generations=2000;
        D=2;
        ch_system=3;

        cout << "Nepodarilo se nacist konfiguracni soubor, evoluce se provede se zakladnimi parametry: 1stDeJong Function, CR=0.8, NP=20, F=0.5, generations=2000, D=2" << endl;
    }

    chaos generator;    //zalozeni noveho generatoru chaosu

    generator.vyberSystem(ch_system); //vyber algoritmu generovani chaosu:
    generator.uloz();  // ulozeni vygenerovanych chaotickych hodnot do souboru a zmena hodnot, aby dosahovaly pouze 0-1



    /* vlastni diferencialni evoluce */
    DE* diferencniEvoluce=new DE(CR, NP, F, generations, D);
    diferencniEvoluce->init(generator, minimum, maximum, testFunction); //nastaveni hodnot DE: pouziteho generatoru, min/max, testovaci funkce a zalozeni uvodni populace

    diferencniEvoluce->evolve(mStrategy, cStrategy);


    delete diferencniEvoluce;           // vymazani objektu diferencialni evoluce (volan destruktor)

    cout << endl << "Pro ukonceni programu zadejte jakykoliv znak: ";
    getchar();

    return 0;
}
