#include "../include/DE.h"
#include "../include/chaos.h"

#include <cstdlib>
#include <iostream>
#include <time.h>
#include <math.h>
#include <cstring>

#define PI  3.14159265
#define E   2.7182818284590452353602874

using namespace std;

int chaosI=0;

DE::DE(double CR, int NP, double F, int generation, int D): crossover(CR), popSize(NP), factor(F), generations(generation), dimension(D){

    this->population=new double[NP * (D+1)];//zalozime populaci o velikosti NP*(D+1) - navyseni o 1 dimenzi je kvuli CostValue
    this->newPop=new double[NP * (D+1)];    //populace potomku
    this->noise=new double[D];              //sumovy vektor jedince nebude mit CostValue
    this->trial=new double[D+1];            //zkusebni vektor jedince CV ma
    this->best=new double[D+1];             //stejne tak i doposud nejlepsi jedinec si uchovava svou nejlepsi hodnotu CV

    //nastavime hodnotu priznaku stagnace. Pokud se hodnota bestCV nezmeni 200<=stagnation<=generations/3, ukoncime evoluci.
    if (generation<=1500)
        maxStagnation=500;
    else
        maxStagnation=generation/3;
    stagnation=0;

    myfile.open("Output.csv");

    if (myfile.is_open()){
        cout << "Textovy vystup naleznete v Output.csv" << endl;
        //ukazka zapisu do myfilemyfile << "1\n";
    }
}

DE::~DE(){
    //destruktor
    if (population) delete population;
    population=0;
    myfile.close();
}

//naplnime populaci zakladnimi hodnotami
void DE::setPopulation(){
    int NP=this->getPopSize();
    int D=this->getDim()+1;     //dimenze+cost value(hodnota ucelove funkce s danymi parametry)

    int bestIndex=0;

    //1. vygeneruju populaci
    for(int i=NP; i<NP*D; i++){
        population[i] = chaosRand(minimum,maximum);
    }

    //2. ohodnotim populaci
    for(int i=0; i<NP; i++){
        population[i]=costFunction(population, i, NP); //vyhodnoceni kvality jedincu v populaci

        /*vyhledam doposud nejlepsi hodnotu Cost Value*/
        if(i==0){//pokud se jedna o prvni vstup, nastavime prvni jako nejlepsi hodnotu
            bestCV=population[i];
        }
        else if(bestCV>population[i]){//jinak porovnavame
            bestCV=population[i];
            bestIndex=i;
        }
    }

    //3. naplnime nejlepsi vektor patricnymi parametry
    for(int i=0; i<=dimension; i++)
        best[i]=population[bestIndex+i*popSize];

        cout << "Nejlepsi CV uvodni populace je " << bestCV << endl;
        myfile << "Nejlepsi CV uvodni populace je " << bestCV << endl;

    return;
}

//inicializacni funkce
void DE::init(chaos generator1, double mini=0.0, double maxi=50.0, int testFnctn=0){
    minimum=mini;
    maximum=maxi;
    testFunction=testFnctn;
    generator=generator1;

    setPopulation();
}

//vypis cele populace a vsech jejich clenu
void DE::getWholePop(int generation=0){
    int NP=this->getPopSize();
    int D=this->getDim();

    //vypis inicializovane populace

    for(int i=1; i<=NP*(D+1); i++){
        myfile << population[i-1] << "; ";
        if (i%NP==0 and i!=1 and i!=NP*(D+1)) {
                myfile << endl ;
        }
    }
    return;
}

//funkce pro kontrolni vypis na konzoli
void DE::getPopulation(int generation=0){
    //vypis nejlepsiho jedince
    cout << endl << "Generace "<< generation << " !";
    myfile << endl <<"Nejlepsi jedinec generace "<< generation << "CV: ";

    for(int ix=0; ix<=dimension; ix++){
        cout << best[ix] << " | ";
        myfile <<" ; " << best[ix];
    }

    stagnation++;

    return;
}

//mutace (vysledkem je nastavni sumoveho vektoru jednoho jedince)
//parametr udava mutacni funkci (viz makra v DE.H)
void DE::evolve(int mutation, int crossbreeding){
    int gen=0;
    for (gen=0; gen<generations && stagnation<maxStagnation; gen++){
        getPopulation(gen+1);

        for (int x=0; x<popSize; x++){

            activeParent=x;//aktivni rodic se vybira v ramci populace postupne jeden po druhem.
            int i=dimension;
            if (crossbreeding==stExp)
                i=selectNoiseIndex()-1;

            switch (mutation){
                case stBest1:
                    best1(i);
                    break;
                case stRand1:
                    rand1(i);
                    break;
                case stRandToBest1:
                    randToBest1(i);
                    break;
                case stBest2:
                    best2(i);
                    break;
                case stRand2:
                    rand2(i);
                    break;
            }

            switch (crossbreeding){
                case stExp:
                    expCross(i);
                    break;
                case stBin:
                    binCross();
                    break;
            }

            for(int ix=0; ix<=dimension; ix++){
                newPop[x+ix*popSize]=trial[ix];   //noveho jedince zaznamename do nove populace
            }
        }
        population=newPop;  //aktualizujeme populaci
    }
    if (stagnation==maxStagnation){
        cout << endl << endl << "Evoluce stagnuje, bude proto ukoncena a to po " << stagnation << " nezlepsujicich se generacich." << endl;
    }
}

//funkce pro inicializaci promennych potrebnych u mutacni fci:
void DE::selectParents(int *r1, int *r2, int *r3, int *r4, int *r5){
    int active=this->activeParent;  //aktivni rodic - rodice, ktereho pouzijeme pro budouci krizeni
    int limit;                      //mezni hodnota populace pro dany algoritmus

	if (r1 && r2){
	    limit=3;
        do {
            *r1=chaosRand(0, popSize);
            *r2=chaosRand(0, popSize);
        } while ((*r2==*r1) || (*r2==active) || (*r1==active));
	}
    if (r3){
        limit=4;
		do {
			*r3 = chaosRand(0, popSize);
		}
		while ((*r3 == active) || (*r3 == *r2) || (*r3 == *r1));
	}

	if (r4){
	    limit=5;
		do {
			*r4 = chaosRand(0, popSize);
		}
		while ((*r4 == active) || (*r4 == *r3) || (*r4 == *r2) || (*r4 == *r1));
	}

	if (r5)	{
		limit=6;
		do {
			*r5 = chaosRand(0, popSize);
		}
		while ((*r5 == active) || (*r5 == *r4) || (*r5 == *r3) || (*r5 == *r2) || (*r5 == *r1));
	}

    if (popSize<limit){
        cout << "Populace je prilis nizka, nez aby se dala vykonat diferencialni evoluce. Musi mit minimalne "<< limit <<" jedince." << endl;
        myfile << "Populace je prilis nizka, nez aby se dala vykonat diferencialni evoluce. Musi mit minimalne "<< limit <<" jedince." << endl;
        return;
    }
}

void DE::best1(int x){
    //vybereme nahodne dva jedince z aktualni populace
    double x1, x2;  //hodnoty nahodne vybranych jedincu
    int r1, r2;     //indexy nahodne vybranych jedincu

    selectParents(&r1, &r2);

    //pro vsechny parametry populace:
    for (int i=0; i<x; i++){
        x1=population[r1+(i+1)*popSize];
        x2=population[r2+(i+1)*popSize];

        noise[i]=best[i+1]+factor*(x1-x2);
    }
}

void DE::rand1(int x){
    //vybereme nahodne tri jedince z aktualni populace
    double x1, x2, x3;  //hodnoty nahodne vybranych jedincu
    int r1, r2, r3;     //indexy nahodne vybranych jedincu

    selectParents(&r1, &r2, &r3);

    //pro vsechny parametry populace:
    for (int i=0; i<x; i++){
        x1=population[r1+(i+1)*popSize];
        x2=population[r2+(i+1)*popSize];
        x3=population[r3+(i+1)*popSize];

        noise[i]=x1+factor*(x2-x3);
    }
}

void DE::randToBest1(int x){
    //vybereme nahodne tri jedince z aktualni populace
    double x1, x2;  //hodnoty nahodne vybranych jedincu
    int r1, r2;     //indexy nahodne vybranych jedincu

    selectParents(&r1, &r2);

    //pro vsechny parametry populace:
    for (int i=0; i<x; i++){
        x1=population[r1+(i+1)*popSize];
        x2=population[r2+(i+1)*popSize];

        noise[i]+=factor*(best[i+1]-noise[i])+factor*(x1-x2);
    }
}

void DE::best2(int x){
    //vybereme nahodne tri jedince z aktualni populace
    double x1, x2, x3, x4;  //hodnoty nahodne vybranych jedincu
    int r1, r2, r3, r4;     //indexy nahodne vybranych jedincu

    selectParents(&r1, &r2, &r3, &r4);

    //pro vsechny parametry populace:
    for (int i=0; i<x; i++){
        x1=population[r1+(i+1)*popSize];
        x2=population[r2+(i+1)*popSize];
        x3=population[r3+(i+1)*popSize];
        x4=population[r4+(i+1)*popSize];

        noise[i]=best[i+1]+factor*(x1+x2-x3-x4);
    }
}

void DE::rand2(int x){
    //vybereme nahodne tri jedince z aktualni populace
    double x1, x2, x3, x4, x5;  //hodnoty nahodne vybranych jedincu
    int r1, r2, r3, r4, r5;     //indexy nahodne vybranych jedincu

    selectParents(&r1, &r2, &r3, &r4, &r5);

    //pro vsechny parametry populace:
    for (int i=0; i<x; i++){
        x1=population[r1+(i+1)*popSize];
        x2=population[r2+(i+1)*popSize];
        x3=population[r3+(i+1)*popSize];
        x4=population[r4+(i+1)*popSize];
        x5=population[r5+(i+1)*popSize];

        noise[i]=x5+factor*(x1+x2-x3-x4);
    }
}

//TODO: u krizeni si zjistit, zda prvni parametr jde automaticky z sumoveho vektoru
//mutace (vysledkem je nastavni sumoveho vektoru jednoho jedince)
void DE::binCross(){

    //nahodnou hodnotu porovname s CR
    //zkusebni vektor naplnime parametrem na zaklade vysledku:
    //nahodna hodnota<CR -> sumovy vektor
    //nahodna hodnota>CR -> parametr 4. rod

        for (int i=1; i<=dimension; i++){
            if (chaosRand(0,1) < crossover){
                if (noise[i-1]<minimum or noise[i-1]>maximum) trial[i]=population[activeParent+i*popSize];
                else trial[i]=noise[i-1];
            }
            else
                trial[i]=population[activeParent+i*popSize];
        }
        //dopocet CostValue vznikleho zkusebniho vektoru:
        double trialEnergy=costFunction(trial, 0, 1);
        double parentEnergy=costFunction(population, activeParent, getPopSize());
        //Vysledny trial vektor pouzijeme, pokud ctvrty rodic ma horsi FitFunkci nezli nove vznikly.
        if (trialEnergy<=parentEnergy)
            trial[0]=trialEnergy;
        else  {
            trial[0]=parentEnergy;
            for (int i=1; i<=dimension; i++){
                trial[i]=population[activeParent+i*popSize];
            }
        }

        if(bestCV>trial[0]){
            //cout << endl << "Nalezl jsem lepsiho jedince, ma CV " << trial[0] << " namisto meho " << bestCV << endl;
            stagnation=0;
            bestCV=trial[0];


            for (int i=0; i<=dimension; i++){
                best[i]=trial[i];
            }
        }
}

void DE::expCross(int noiseIndex){

    //nahodnou hodnotu porovname s CR
    //dokud plati ze:
    //nahodna hodnota<CR -> sumovy vektor
    //nahodna hodnota>CR -> vsechny ostatni parametry 4. rod

    //nahodna hodnota=noiseIndex

    for (int i=1; i<=dimension; i++){

        if (i<=noiseIndex){
            //osetreni abychom nevylezli mimo meze
            if (noise[i-1]<minimum or noise[i-1]>maximum) trial[i]=population[activeParent+i*popSize];
            else trial[i]=noise[i-1];
        }
        else
            trial[i]=population[activeParent+i*popSize];

    }

        //dopocet CostValue vznikleho zkusebniho vektoru:
        double trialEnergy=costFunction(trial, 0, 1);
        double parentEnergy=costFunction(population, activeParent, getPopSize());//CV rodice


        //Vysledny trial vektor pouzijeme, pokud ctvrty rodic ma horsi FitFunkci nezli nove vznikly.
        if (trialEnergy<=parentEnergy)
            trial[0]=trialEnergy;
        else  {
            trial[0]=parentEnergy;
            for (int i=1; i<=dimension; i++){
                trial[i]=population[activeParent+i*popSize];
            }
        }


        if(bestCV>trial[0]){
            stagnation=0;
            bestCV=trial[0];

            for (int i=0; i<=dimension; i++){
                best[i]=trial[i];
            }
        }
}

int DE::selectNoiseIndex(){

    int noiseIndex;

    for (noiseIndex=0; (noiseIndex<=dimension) && (chaosRand(0,1) < crossover); noiseIndex++)
    {}

    return noiseIndex;
}

double DE::costFunction(double * testPopulation, int index, int popSiz){
    double costValue=0;

    //0 - 1st deJong - ucelova funkce je SUMA x^2
    if (testFunction==0){
        for (int i=1; i<=dimension; i++){
            double testValue=testPopulation[index+i*popSiz];
            costValue+=(testValue*testValue);
        }
    }

    //1 - 2nd deJong - Rosenbrockovo sedlo - ucelova funkce je SUMA 100 (x(i)^2 - x(i+1))^2 + (1-x(i))^2
    else if (testFunction==1){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];//pristi parametr stejneho jedince
            costValue+=100*(x1*x1-x2)*(x1*x1-x2)+((1-x1)*(1-x1));
        }
    }

    //2 - 3rd deJong - ucelova funkce je SUMA abs(x)
    else if (testFunction==2){
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            if (x1<0) x1*=-1; //absolutni hodnota
            costValue+=x1;
            //cout << "x1=" << x1 << " x2=" << x2 << endl;
        }
    }

    //3 - 4th deJong - ucelova funkce je SUMA i (x)^4
    else if (testFunction==3){
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            costValue+=i*x1*x1*x1*x1;
        }
    }

    //4 - Rastriginova funkce - ucelova funkce je 10D SUMA (x)^2 - 10 cos(2 PI x)
    else if (testFunction==4){
        costValue=10*dimension;
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            //costValue+=10*dimension*(x1*x1-10*cos(2*PI*x1));
            costValue+=(x1*x1-10*cos(2*PI*x1));
        }
    }

    //5 - Schwefelova funkce - ucelova funkce je SUMA -(x) sin (abs(x))
    else if (testFunction==5){
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double absx1=x1;
            if (absx1<0) absx1*=-1;
            costValue-=(x1*sin(sqrt(absx1)));
        }
    }

    //6 - Griewangkova funkce - ucelova funkce je 1 + SUMA x^2/4000 - PRODUKT cos (x/sqrt(i))
    else if (testFunction==6){
            double costSum=0;
            double costProduct=1;
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            costSum+=(x1*x1)/4000;
            costProduct*=cos(x1/sqrt(i));
        }
        costValue=1+costSum-costProduct;
    }

    //7 - Sinova obalkova sinusoidalni funkce - ucelova funkce je -SUMA (0.5+(sin(?sqrt(x(i)^2+x(i+1)^2)-0.5)^2)/(1+0.001(x(i)^2+x(i+1)^2))^2)
    //vysledek urcuje kruznice v En
    else if (testFunction==7){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double pyth=x1*x1+x2*x2;//pythagoruv soucet ctvercu

            costValue+=0.5+((pow(sin(sqrt(pyth)-0.5), 2))/pow(1+0.001*(pyth), 2));
        }
        costValue*=-1;
    }

    //8 - Roztazena sinusoidalni V funkce - ucelova funkce je SUMA ((x(i)^2+x(i+1)^2)^(1/4) * (sin(50*(x(i)^2+x(i+1)^2)^(1/10)^2 ) + 1)
    else if (testFunction==8){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double pyth=x1*x1+x2*x2;//pythagoruv soucet ctvercu

            costValue+=(pow(pyth, 0.25)*(pow(sin(50*pow(pyth, 0.1)), 2)+1));
        }
    }

    //9 - 1. Ackleyho funkce - ucelova funkce je SUMA (1/e^(1/5) sqrt(x(i)^2+x(i+1)^2) + 3 cos (2x(i)) + sin (2x(x+1)))
    //-30-30
    else if (testFunction==9){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double pyth=x1*x1+x2*x2;//pythagoruv soucet ctvercu

            costValue+=( sqrt(pyth)/pow(E, 0.2) + 3*(cos(2*x1) + sin(2*x2)));
        }
    }

    //10 - 2. Ackleyho funkce
    else if (testFunction==10){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double pyth=x1*x1+x2*x2;//pythagoruv soucet ctvercu

            //costValue+=( 20 + E - (20/pow(E, 0.2*(sqrt(pyth/2)))) - pow(E, 0.5*(cos(2*PI*x1)+cos(2*PI*x2))) );//dle literatury
            //costValue+=( 20 + pow(E, -20) * pow(E, -0.2*(sqrt(pyth/2))) - pow(E, 0.5*(cos(2*PI*x1)+cos(2*PI*x2))) );//dle internetoveho zdroje http://www.maths.uq.edu.au/
            costValue+=( 20 + E - pow(E, (cos(2*PI*x1)+cos(2*PI*x2))/2) - 20*pow(E, -0.2*(sqrt(pyth/2))));      //z internetu http://tracer.lcc.uma.es/problems/ackley/ackley.html
            //vybrana funkce koresponduje s literaturou, ovsem ma lepsi vysledky.
        }
    }

    //11 - Plato vajec
    else if (testFunction==11){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double abs1=x1-x2-47;
            double abs2=x2+47+x1/2;

            if (abs1<0) abs1*=-1;
            if (abs2<0) abs2*=-1;

            costValue+=(-x1 * sin(sqrt(abs1)) - (x2+47)*sin(sqrt(abs2)));
        }
    }

    //12 - Ranova funkce
    else if (testFunction==12){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double abs1=x2+1-x1;
            double abs2=x2+1+x1;

            if (abs1<0) abs1*=-1;
            if (abs2<0) abs2*=-1;

            costValue+=( x1*sin(sqrt(abs1))*cos(sqrt(abs2)) + (x2+1)*cos(sqrt(abs1))*sin(sqrt(abs2)) );
            //hodnotu glob. minima najdeme na http://www.cs.unm.edu/~neal.holts/GECCO2013/benchmarkFunction/rana.html
        }
    }

    //13 - Patologicka funkce
    //!neznam minimum, nemohu otestovat
    else if (testFunction==13){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            costValue+=( 0.5 + (pow(sin(sqrt(100*x1*x1-x2*x2)), 2)-0.5) / (1 + 0.001*pow(x1*x1-2*x1*x2+x2*x2, 2)) );//dle literatury
        }
    }

    //14 - Michalewiczova funkce
    else if (testFunction==14){
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            //double x2=testPopulation[index+(i+1)*popSiz];

            //costValue+=(sin(x1)*pow(sin(x1*x1/PI), 20) + sin(x2)*pow(sin(2*x2*x2/PI), 20) );//dle literatury, je treba zmenit podminku do i<dimension
            //costValue+=( sin(x1)*pow(sin(x2*x2/PI), 20) + sin(x2)*pow(sin(2*x2*x2/PI), 20) );//z maths.edu.
            costValue+=sin(x1)*pow(sin(i*x1*x1/PI), 20);//jednodussi zapis fce pri jinem omezeni sumy
        }
        costValue*=-1;
    }

    //15 - Mastersova funkce
    else if (testFunction==15){
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];
            double x2=testPopulation[index+(i+1)*popSiz];

            double x=x1*x1+x2*x2+0.5*x1*x2;

            costValue-=pow(E, -x/8)*cos(4*sqrt(x));
        }
    }

    //100 - Nova uzivatelska funkce
    else if (testFunction==100){

        //pro jeden parametr:
        for (int i=1; i<=dimension; i++){
            double x1=testPopulation[index+i*popSiz];//parametr x

            costValue+=0*x1;//zapis nove funkce patri sem
        }
        //pro vice parametru:
        for (int i=1; i<dimension; i++){
            double x1=testPopulation[index+i*popSiz];//parametr 1
            double x2=testPopulation[index+(i+1)*popSiz];//parametr 2

            costValue+=0*x1+0*x2;//zapis nove funkce patri sem
        }
    }

    return costValue;

}

//generovani nahodneho cisla v intervalu od minima do maxima
double DE::chaosRand(double minimum=0, double maximum=1){
    if(chaosI > MAX_DELKA-1)
    {
        generator.reev();
        chaosI=0;
    }

    chaosI++;
    //return minimum + (fabs(generator.getX(chaosI))/generator.getMax()) * (maximum-minimum);
     return minimum + (generator.getX(chaosI) * (maximum-minimum));
}
